#!/usr/bin/perl
#
# Jonathan Klassen
#
# ORFcor.pl
# v1.303 - May 7, 2013
#
# Script for correcting multiple fasta files for misannotated start and termination sites and chimeric assemblies, based on homology to closely related sequences
# Multithreaded, one thread for each multiple alignment considered

use strict;
use warnings;
use Parallel::ForkManager;
use Getopt::Long;

######################################################################################################################
# Processes input arguments
######################################################################################################################

my $usage = "ORFcor.pl

DEPENDANCIES: BLAST+ suite, MUSCLE, Perl \"Parallel::ForkManager\" module, EMBOSS \"transeq\" if using nucleotides as input data

USAGE:
-a	length difference to the reference sequences to start 5' trimming or Xs	DEFAULT: 5			e.g., perl ORFcor.pl -a 10
-b	length difference to the reference sequences to start 3' trimming or Xs	DEFAULT: 20			e.g., perl ORFcor.pl -b 20
-c	set number of cores							DEFAULT: 1			e.g., perl ORFcor.pl -c 8
-d	sequence ID allowed for the bulk of the sequence (for AAI correction)	DEFAULT: 0.9			e.g., perl ORFcor.pl -d 0.75
-e 	set max e-value for BLAST						DEFAULT: 1e-5			e.g., perl ORFcor.pl -e 1e-10
-f	length difference to the reference sequences to declare a 5' chimera	DEFAULT: 10			e.g., perl ORFcor.pl -f 20
-g	length difference to the reference sequences to declare a 3' chimera	DEFAULT: 30			e.g., perl ORFcor.pl -g 20	
-h	displays this usage statement (also using --help)							e.g., perl ORFcor.pl -h 
-i 	input list of paths to multiple sequence files for each hmm type	DEFAULT: Mult_hmmscan_faa.list	e.g., perl ORFcor.pl -i inlist
-j	minimum number of closely reference seqs needed for correction		DEFAULT: 5			e.g., perl ORFcor.pl -j 3
-k	maximum number of (non-identical) sequences used for correction		DEFAULT: 1000			e.g., perl ORFcor.pl -k 5
-l	maximum number of sequences retained by BLAST (must be > 2)		DEFAULT: 1000			e.g., perl ORFcor.pl -l 12
-p	input sequences are protein (Y or N; runs translation module if N)	DEFAULT: Y			e.g., perl ORFcor.pl -p N
-q	run quietly, i.e., no STDOUT (Y or N)					DEFAULT: N			e.g., perl ORFcor.pl -q Y
-r	remove intermediate output files (Y or N)				DEFAULT: Y			e.g., perl ORFcor.pl -r N

OUTPUT FILES:
	ORFcor_summary.out					Table listing the numbers of each type of correction for each sequence type
	ORFcor_details.out					A list of the corrected sequences and the amount of each sequence corrected
	ORFcor_individual_sequence_files/<sequence type>/	Directory tree containing individual sequence files, their BLASTp results and corrected versions
	Mult_corrected_faa.list					A list of the multiple fasta files corrected by ORFcor
	ORFcor_multiple_sequence_files/				Directory containing concatenated corrected multiple fasta files
	ORFcor_individual_sequence_files_nucl/			Directory tree containing individual ffn files
	ORFcor_multiple_sequence_files_nucl/			Directory containing corrected multiple ffn files
	mult_faas_translated_from_nucl/				Directory containing faa files translated from ffn files
";

# input arguments

my %options = ();
GetOptions(
	"a=i" => \$options{min_5_trimming_length},		"b=i" => \$options{min_3_trimming_length},
	"c=i" => \$options{cores}, 				"d=s" => \$options{min_ID},
	"e=s" => \$options{evalue},				"f=s" => \$options{min_5_chimera_length},
	"g=s" => \$options{min_3_chimera_length},		"h" => \$options{help},
	"help" => \$options{help}, 				"i=s" => \$options{infile},
	"j=i" => \$options{min_refs},				"k=i" => \$options{max_refs},
	"l=i" => \$options{max_BLAST_seqs},			"p=s" => \$options{protein},
	"q=s" => \$options{quiet},				"r=s" => \$options{remove_intermediates}
);

# display usage statement if called
 
die $usage if ($options{help});

# default arguments

unless ($options{min_5_trimming_length}){ $options{min_5_trimming_length} = 5 }
unless ($options{min_3_trimming_length}){ $options{min_3_trimming_length} = 20 }
unless ($options{cores}){ $options{cores} = 1 }
unless ($options{min_ID}){ $options{min_ID} = 0.9 }
unless ($options{evalue}){ $options{evalue} = 1e-5 }
unless ($options{min_5_chimera_length}){ $options{min_5_chimera_length} = 10 }
unless ($options{min_3_chimera_length}){ $options{min_3_chimera_length} = 30 }
unless ($options{infile}){ $options{infile} = "Mult_hmmscan_faa.list" }
unless ($options{min_refs}){ $options{min_refs} = 5 }
unless ($options{max_refs}){ $options{max_refs} = 1000 }
unless ($options{max_BLAST_seqs}){ $options{max_BLAST_seqs} = 1000 }
unless ($options{protein}){ $options{protein} = "Y" }
unless ($options{quiet}){ $options{quiet} = "N" }
unless ($options{remove_intermediates}){ $options{remove_intermediates} = "Y" }

# mystery input flags not allowed

die "Unrecognized command line arguments: @ARGV\n" if ($ARGV[0]);

# checks correct parameter formatting

die "Unrecognized command line arguments: -e = $options{evalue}\n$usage" unless ($options{evalue} =~ /^\d+(\.?\d+)?$/ 
	or $options{evalue} =~ /^\d+e-\d+$/);
die "Unrecognized command line arguments: -q = $options{quiet}\n$usage" unless ($options{quiet} eq "Y" or $options{quiet} eq "N");
die "Unrecognized command line arguments: -r = $options{remove_intermediates}\n$usage"
	unless ($options{remove_intermediates} eq "Y" or $options{remove_intermediates} eq "N");

# checks if input and output files exist

unless (-e $options{infile}){ die "$options{infile} does not exist\n$usage" }
unless ($options{quiet} eq "Y"){
	if (-e "ORFcor_details.out"){ print "WARNING: ORFcor_details.out exists and has been overwritten\n" }
	if (-d "ORFcor_individual_sequence_files"){ 
		print "WARNING: ORFcor_individual_sequence_files output directory exists and has been overwritten\n";
	}
	if (-e "Mult_corrected_faa.list"){ print "WARNING: Mult_corrected_faa.list exists and has been overwritten\n" }
	if (-d "ORFcor_multiple_sequence_files"){ print "WARNING: ORFcor_multiple_sequence_files output directory exists and has been overwritten\n" }
	if (-e "ORFcor_summary.out"){ print "WARNING: ORFcor_summary.out exists and has been overwritten\n" }
	if ($options{protein} eq "N"){
		if (-d "ORFcor_individual_sequence_files_nucl"){ 
			print "WARNING: ORFcor_individual_sequence_files_nucl exists and has been overwritten\n";
		}
		if (-d "ORFcor_multiple_sequence_files_nucl"){
			print "WARNING: ORFcor_multiple_sequence_files_nucl exists and has been overwritten\n";
		}
		if (-d "mult_faas_translated_from_nucl"){ print "WARNING: mult_faas_translated_from_nucl exists and has been overwritten\n" }
	}
}
if (-e "ORFcor_details.out"){ system "rm ORFcor_details.out" }
if (-d "ORFcor_individual_sequence_files"){ system "rm -rf ORFcor_individual_sequence_files" }
if (-e "Mult_corrected_faa.list"){ system "rm Mult_corrected_faa.list" }
if (-d "ORFcor_multiple_sequence_files"){ system "rm -rf ORFcor_multiple_sequence_files" }
if (-e "ORFcor_summary.out"){ system "rm ORFcor_summary.out" }
if ($options{protein} eq "N"){
	if (-d "ORFcor_individual_sequence_files_nucl"){ system "rm -rf ORFcor_individual_sequence_files_nucl" }
	if (-d "ORFcor_multiple_sequence_files_nucl"){ system "rm -rf ORFcor_multiple_sequence_files_nucl" }
	if (-d "mult_faas_translated_from_nucl"){ system "rm -rf mult_faas_translated_from_nucl exists and has been overwritten" }
}

print "
----------------------------------------------------------------
ORFcor.pl	Jonathan Klassen 	v1.302: Jan. 25, 2013

parameters used:
	minimum sequence length at 5' end deviating from the mean reference values required for correction = $options{min_5_trimming_length}
	minimum sequence length at 3' end deviating from the mean reference values required for correction = $options{min_3_trimming_length}
	number of cores to use = $options{cores}
	minimum id between query and accepted reference sequence = $options{min_ID}
	minimum e-value set for BLASTp = $options{evalue}
	minimum sequence length at 5' end deviating from the mean reference values required for correction as a chimera = $options{min_5_chimera_length}
	minimum sequence length at 3' end deviating from the mean reference values required for correction as a chimera = $options{min_3_chimera_length}
	input list of multiple sequence files, one for each protein type = $options{infile}
	minimum number of reference sequences required for correction = $options{min_refs}
	maximum number of reference sequences used for correction = $options{max_refs}
	maximum number of blast hits retained = $options{max_BLAST_seqs}
	input is protein sequence = $options{protein}
	run quietly = $options{quiet}
	remove intermediate files = $options{remove_intermediates}
----------------------------------------------------------------
" if ($options{quiet} eq "N");

# makes output directories

mkdir "ORFcor_individual_sequence_files";
mkdir "ORFcor_multiple_sequence_files";
if ($options{protein} eq "N"){ 
	mkdir "ORFcor_individual_sequence_files_nucl";
	mkdir "ORFcor_multiple_sequence_files_nucl";
	mkdir "mult_faas_translated_from_nucl";
}


######################################################################################################################
# Loads input files, derives names from them
######################################################################################################################

open (INFILELIST, $options{infile}) or die "Cannot open $options{infile}";

my (@input_file_list, @input_file_names);
while (<INFILELIST>){
	chomp;
	next unless ($_ =~ /\w/);	
	push @input_file_list, $_;
	my $temp;
	if (/\/?((\w|-)+)\..*$/){ $temp = $1 }
	elsif (/\/?((\w|-)+)$/){ $temp = $1 }
	$temp =~ s/.*\///;
	$temp =~ s/mult_//;
	push @input_file_names, $temp;
}

##############################################################################
# Multithreaded running of ORFcor algorithm
# Each forked process is stored in the temp_outputs directory and collated at the end
##############################################################################

sub seq_corrector($);

our %summary_output;

my $pm = new Parallel::ForkManager($options{cores});

for my $a (0..$#input_file_list){
	$pm->start and next;
	my @return = seq_corrector($a);

	unless (-d "temp_outputs"){ mkdir "temp_outputs" }
	open (TEMPOUT, ">temp_outputs/$a.tempout") or die "Cannot open temp_outputs/$a.tempout";
	print TEMPOUT join "\t", @return;
	close TEMPOUT;
	$pm->finish;
}
$pm->wait_all_children();

for my $a (0..$#input_file_list){
	open (TEMPIN, "temp_outputs/$a.tempout") or die "Cannot open temp_outputs/$a.tempout";
	my @return;
	while (<TEMPIN>){
		@return = split "\t";
	}

	##############################################################################
	# parses returned summary data
	##############################################################################

	my $temp_counter = shift @return;

	# initialize counters for each correction type

	$summary_output{$input_file_names[$temp_counter]}{five_chimera} = $summary_output{$input_file_names[$temp_counter]}{five_long} = 
		$summary_output{$input_file_names[$temp_counter]}{five_short} = $summary_output{$input_file_names[$temp_counter]}{three_chimera} = 
		$summary_output{$input_file_names[$temp_counter]}{three_long} = $summary_output{$input_file_names[$temp_counter]}{three_short} = 0;

	# parses corrected sequences returned from the subroutine

	foreach (@return){
		my @returned_sequence = split ",";
		if ($returned_sequence[1] != 0){
			$summary_output{$input_file_names[$temp_counter]}{five_chimera}++;
			push @{$summary_output{$input_file_names[$temp_counter]}{five_chimeras}}, "$returned_sequence[0]:$returned_sequence[1]";
		}
		if ($returned_sequence[2] != 0){
			$summary_output{$input_file_names[$temp_counter]}{five_long}++;
			push @{$summary_output{$input_file_names[$temp_counter]}{five_longs}}, "$returned_sequence[0]:$returned_sequence[2]";
		}
		if ($returned_sequence[3] != 0){
			$summary_output{$input_file_names[$temp_counter]}{five_short}++;
			push @{$summary_output{$input_file_names[$temp_counter]}{five_shorts}}, "$returned_sequence[0]:$returned_sequence[3]";
		}
		if ($returned_sequence[4] != 0){
			$summary_output{$input_file_names[$temp_counter]}{three_chimera}++;
			push @{$summary_output{$input_file_names[$temp_counter]}{three_chimeras}}, "$returned_sequence[0]:$returned_sequence[4]";
		}
		if ($returned_sequence[5] != 0){
			$summary_output{$input_file_names[$temp_counter]}{three_long}++;
			push @{$summary_output{$input_file_names[$temp_counter]}{three_longs}}, "$returned_sequence[0]:$returned_sequence[5]";
		}
		if ($returned_sequence[6] != 0){
			$summary_output{$input_file_names[$temp_counter]}{three_short}++;
			push @{$summary_output{$input_file_names[$temp_counter]}{three_shorts}}, "$returned_sequence[0]:$returned_sequence[6]";
		}
	}
}
system "rm -rf temp_outputs";

######################################################################################################################
# Creates summary output files
######################################################################################################################

open (OUTSUMMARY, ">ORFcor_summary.out") or die "Cannot open ORFcor_summary.out";
open (OUTDETAILS, ">ORFcor_details.out") or die "Cannot open ORFcor_details.out";
print OUTSUMMARY "Protein_type\tFive_prime_chimeras\tFive_prime_too_long\tFive_prime_too_short\tThree_prime_chimeras\tThree_prime_too_long\tThree_prime_too_short\n";

foreach my $hmm_type (@input_file_names){
	print OUTSUMMARY "$hmm_type\t$summary_output{$hmm_type}{five_chimera}\t$summary_output{$hmm_type}{five_long}\t",
		"$summary_output{$hmm_type}{five_short}\t$summary_output{$hmm_type}{three_chimera}\t$summary_output{$hmm_type}{three_long}",
		"\t$summary_output{$hmm_type}{three_short}\n";
	unless (!$summary_output{$hmm_type}{five_chimeras}[0] and !$summary_output{$hmm_type}{five_longs}[0] 
		and !$summary_output{$hmm_type}{five_shorts}[0] and !$summary_output{$hmm_type}{three_chimeras}[0] 
		and !$summary_output{$hmm_type}{three_longs}[0] and !$summary_output{$hmm_type}{three_shorts}[0]){
		print OUTDETAILS "\n# $hmm_type\n";
	}
	unless (!$summary_output{$hmm_type}{five_chimeras}[0]){
		foreach (@{$summary_output{$hmm_type}{five_chimeras}}){
			my @temp = split ":";
			print OUTDETAILS "Sequence having 5' chimeras: $temp[0]\t$temp[1] amino acids replaced with \"X\"'s\n";
		}
	}
	unless (!$summary_output{$hmm_type}{five_longs}[0]){
		foreach (@{$summary_output{$hmm_type}{five_longs}}){
			my @temp = split ":";
			print OUTDETAILS "Sequence having long 5' ends: $temp[0]\t$temp[1] amino acids trimmed\n";
		}
	}
	unless (!$summary_output{$hmm_type}{five_shorts}[0]){
		foreach (@{$summary_output{$hmm_type}{five_shorts}}){
			my @temp = split ":";
			print OUTDETAILS "Sequence having short 5' ends: $temp[0]\t$temp[1] amino acid \"X\"'s added\n";
		}
	}
	unless (!$summary_output{$hmm_type}{three_chimeras}[0]){
		foreach (@{$summary_output{$hmm_type}{three_chimeras}}){
			my @temp = split ":";
			print OUTDETAILS "Sequence having 3' chimeras: $temp[0]\t$temp[1] amino acids replaced with \"X\"'s\n";
		}
	}
	unless (!$summary_output{$hmm_type}{three_longs}[0]){
		foreach (@{$summary_output{$hmm_type}{three_longs}}){
			my @temp = split ":";
			print OUTDETAILS "Sequence having long 3' ends: $temp[0]\t$temp[1] amino acids trimmed\n";
		}
	}
	unless (!$summary_output{$hmm_type}{three_shorts}[0]){
		foreach (@{$summary_output{$hmm_type}{three_shorts}}){
			my @temp = split ":";
			print OUTDETAILS "Sequence having short 3' ends: $temp[0]\t$temp[1] amino acid \"X\"'s added\n";
		}
	}
}
close OUTSUMMARY;
close OUTDETAILS;

######################################################################################################################
# Create list of corrected fasta files 
######################################################################################################################

open (OUTLIST, ">Mult_corrected_faa.list") or die "Cannot open Mult_corrected_faa.list";
foreach (@input_file_names){
	next if (/^\./);
	print OUTLIST "ORFcor_multiple_sequence_files/$_\_corrected.faa\n";
}

######################################################################################################################
# Cleans up extra files if running quietly 
######################################################################################################################

if ($options{remove_intermediates} eq "Y"){
	system "rm -rf ORFcor_individual_sequence_files";
	if ($options{protein} eq "N"){
		system "rm -rf ORFcor_individual_sequence_files";
		system "rm -rf ORFcor_multiple_sequence_files";
		system "rm -rf ORFcor_individual_sequence_files_nucl";
		system "rm -rf mult_faas_translated_from_nucl";
		system "rm Mult_corrected_faa.list";
	}
}

######################################################################################################################
# Detecting and correcting likely incorrect sequences
######################################################################################################################

sub seq_corrector ($){
	my $counter = $_[0]; # index counter for global vars
	my @return; # return array for sub

	if ($options{quiet} eq "N"){ print "Correcting: $input_file_names[$counter]\n" }

	######################################################################################################################
	# Converts nucleotide sequences to protein sequences, if required
	# uses TRANSEQ in the EMBOSS package for nucleotide translation
	# requires that all nucleotides are in the +1 reading frame, and that there are no bases outside of the coding region
	# if the last codon of the gene is a stop codon, this is recorded but omitted from the protein sequence
	######################################################################################################################

	if ($options{protein} eq "N"){

		system "transeq -frame=1 -sequence=$input_file_list[$counter] -outseq=mult_faas_translated_from_nucl/$input_file_names[$counter].faa -auto";

		# checks for stop codons in translation

		open (INFILE, "mult_faas_translated_from_nucl/$input_file_names[$counter].faa") or die;
		open (TEMPOUT, ">mult_faas_translated_from_nucl/tempout_$counter") or die;

		my ($seq, $name, @filenames, @temp);
		while (<INFILE>){
			chomp;
			if (/^>(\w+)/){
				my $tempname = $1;
				if ($seq and $seq =~ /\*/){ # checks for bad translations before next new sequence
					@temp = split '', $seq; 
					if ($temp[$#temp] =~ /\*/){ pop @temp }
					die "ERROR: $name in $input_file_list[$counter] could not be translated correctly\n",@temp,"\n" 
						if (grep {/\*/} @temp);
					print TEMPOUT ">$name\n", join "", @temp, "\n";
				}
				elsif ($seq){
					@temp = split '', $seq; 
					print TEMPOUT ">$name\n", join "", @temp, "\n";
				}
				($name = $tempname) =~ s/_1$//;
				push @filenames, $name;
				$seq = '';
			}
			else {	
				$seq = $seq.$_;
			}
		}

		@temp = split '', $seq; 
		if ($seq and $seq =~ /\*/){ # checks for bad translation of the last sequence
			if ($temp[$#temp] =~ /\*/){ pop @temp }
			die "ERROR: $name in $input_file_list[$counter] could not be translated correctly\n",@temp,"\n" if (grep {/\*/} @temp);
		}
		print TEMPOUT ">$name\n", join "", @temp, "\n";
		@temp = ();
		close INFILE;
		system "mv mult_faas_translated_from_nucl/tempout_$counter mult_faas_translated_from_nucl/$input_file_names[$counter].faa";

		# makes directory of individual nucleotide files 

		my $found;
		mkdir "ORFcor_individual_sequence_files_nucl/$input_file_names[$counter]";
		open (INFILE, $input_file_list[$counter]) or die "Cannot open $input_file_list[$counter]";
		while (<INFILE>){
			chomp;
			if (/>/){
				my $line = $_;
				$found = '';
				foreach (@filenames){
					if ($line =~ /$_/){
						open (OUTFILE, ">ORFcor_individual_sequence_files_nucl/$input_file_names[$counter]/$_.ffn") or 
							die "ORFcor_individual_sequence_files_nucl/$input_file_names[$counter]/$_.ffn";
						print OUTFILE $line, "\n";
						$found = "y";
					}
				}	
				die "Cannot find filename: @filenames\n$line" if ($found ne "y");
			}
			else {
				s/\*//g;
				print OUTFILE $_, "\n";
			}
		}
					
	}

	# parses protein files

	$return[0] = $counter;

	# splits input multiple fasta file into individual files
	#	rationale: large files could have too large a RAM overhead during parsing with multithreading

	mkdir "ORFcor_individual_sequence_files/$input_file_names[$counter]";
	if ($options{protein} eq "Y"){ open (INFILE, $input_file_list[$counter]) or die "Cannot open $input_file_list[$counter]" }
	elsif ($options{protein} eq "N"){ open (INFILE, "mult_faas_translated_from_nucl/$input_file_names[$counter].faa") 
		or die "mult_faas_translated_from_nucl/$input_file_names[$counter].faa" }

	while (<INFILE>){
		if (/^>((\w|-)+)/){
			open (OUTFILE, ">ORFcor_individual_sequence_files/$input_file_names[$counter]/$1.faa") 
				or die "Cannot open ORFcor_individual_sequence_files/$input_file_names[$counter]/$1.faa";
			print OUTFILE $_;
		}
		else {
			print OUTFILE;
		}
	}
	close OUTFILE;

	# Lists input fasta files

	opendir (DIR, "ORFcor_individual_sequence_files/$input_file_names[$counter]");
	my @individual_faas = readdir (DIR);

	######################################################################################################################
	# BLASTs each individual faa file vs the multiple faa file
	# Determines unaligned regions for each sequence and counts each error type in set
	# Counting necessary before correction to identify cases where declaring short vs long is equivocal
	######################################################################################################################

	# Format multiple fasta file for BLASTing against

	if ($options{protein} eq "Y"){
		system "makeblastdb -in $input_file_list[$counter] -logfile $input_file_list[$counter].makeblastdb.log -dbtype prot";
		system "rm -f $input_file_list[$counter].makeblastdb.log";
	}
	elsif ($options{protein} eq "N"){
		system "makeblastdb -in mult_faas_translated_from_nucl/$input_file_names[$counter].faa -logfile $input_file_names[$counter].makeblastdb.log -dbtype prot";
		system "rm -f $input_file_list[$counter].makeblastdb.log";
	}

	# BLAST each individual_faa vs the multiple fasta file

	my (%total_hit_counter, %unaligned, %consensus, $total_five_chimera, $total_five_long, $total_five_short, $total_three_chimera,
		$total_three_long, $total_three_short); # vars tallying changes for second loop
	my $no_hits = 0; # var counting outgroup seqs, to exclude from ambiguous calculation
	foreach my $individual_faa (@individual_faas){
		next if ($individual_faa =~ /^\./);

		my %hits;

		# BLASTs each protein vs all others of that hmm type, takes the top $options{max_BLAST_seqs} hits (excluding self) < $expect

		if ($options{protein} eq "Y"){
			system "blastp -query ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa -db $input_file_list[$counter] -evalue $options{evalue} -max_target_seqs $options{max_BLAST_seqs} -comp_based_stats F -outfmt \"7 qseqid sseqid qlen slen qstart qend sstart send length evalue bitscore nident positive\" -out ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa.blast";
		}
		elsif ($options{protein} eq "N"){
			system "blastp -query ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa -db mult_faas_translated_from_nucl/$input_file_names[$counter].faa -evalue $options{evalue} -max_target_seqs $options{max_BLAST_seqs} -comp_based_stats F -outfmt \"7 qseqid sseqid qlen slen qstart qend sstart send length evalue bitscore nident positive\" -out ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa.blast";
		}

		# parse output blastp file

		open (INFILE, "ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa.blast") 
			or die "Cannot open ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa.blast";
		my ($query_name, $query_length);
		while (<INFILE>){
			chomp;
			next if (/^#/);
			my @line = split "\t";

			# only keeps hits up to a threshold, unless 100% identical
			# necessary because 100% identical but overextended reference sequences will have higher BLAST scores than 
			# non-overextended ones, and therefore artifically ranked higher

			next if (scalar keys %hits > $options{max_refs} and $line[11] != $line[8]);

			$line[0] =~ /^((\w|-)+)/;
			$query_name = $1;
			$query_length = $line[2];
			my $hit_name = $line[1];
			next if ($hits{$hit_name}); # only single best aligned region retained;
			%{$hits{$hit_name}} = (
				hit_length => $line[3],
				identities => $line[11],
				aligned => $line[8],
				query_start => $line[4],
				query_end => $line[5],
				hit_start => $line[6],
				hit_end => $line[7]
			);
			$hits{$hit_name}{'percent_identities'} = $line[11] / $line[8];
		}
		close INFILE;

		# only keeps hits above identities threshold

		foreach my $hit_name (keys %hits){
			if ($hits{$hit_name}{'percent_identities'} < $options{min_ID}){
				delete $hits{$hit_name};
			}
		}

		# determines how much of each end of the query and subject are unaligned

		my (@five_query_unaligned, @five_hit_unaligned, @three_query_unaligned, @three_hit_unaligned);
		my $hit_counter = 0;

		foreach my $hit_name (sort keys %hits){
			next if ($hit_name =~ /$query_name/);
			next if ($hit_name eq '');

			push @five_query_unaligned, $hits{$hit_name}{'query_start'} - 1;
			push @five_hit_unaligned, $hits{$hit_name}{'hit_start'} - 1;
			push @three_query_unaligned, $query_length - $hits{$hit_name}{'query_end'};
			push @three_hit_unaligned, $hits{$hit_name}{'hit_length'} - $hits{$hit_name}{'hit_end'};
			$hit_counter++;
		}

		# requires at least $options{min_refs} genomes to compare with the query

		if ($hit_counter < $options{min_refs}){ 
			$no_hits++; 
			next;
		} 

		# for each hit, determines whether it implies 5' and/or 3' long, short or chimera from the number of unaligned positions

		$total_hit_counter{$query_name} = $hit_counter;
		for my $a (0..$#five_query_unaligned){ # by definition all unaligned arrays are the same length
			if ($five_query_unaligned[$a] >= $options{min_5_chimera_length} and $five_hit_unaligned[$a] >= 
				$options{min_5_chimera_length}){
				$unaligned{$query_name}{five_chimera_query_unaligned_freqs}{$five_query_unaligned[$a]}++;
				$unaligned{$query_name}{five_chimera_hit_unaligned_freqs}{$five_hit_unaligned[$a]}++;
				$unaligned{$query_name}{five_chimera_count}++;
			}
			elsif ($five_query_unaligned[$a] - $five_hit_unaligned[$a] >= $options{min_5_trimming_length}){
				$unaligned{$query_name}{five_long_unaligned_freqs}{$five_query_unaligned[$a]}++;
				$unaligned{$query_name}{five_long_count}++;
			}
			elsif ($five_hit_unaligned[$a] - $five_query_unaligned[$a] >= $options{min_5_trimming_length}){
				$unaligned{$query_name}{five_short_unaligned_freqs}{$five_hit_unaligned[$a]}++;
				$unaligned{$query_name}{five_short_count}++;
			}
			if ($three_query_unaligned[$a] >= $options{min_3_chimera_length} and $three_hit_unaligned[$a] >= 
				$options{min_3_chimera_length}){
				$unaligned{$query_name}{three_chimera_query_unaligned_freqs}{$three_query_unaligned[$a]}++;
				$unaligned{$query_name}{three_chimera_hit_unaligned_freqs}{$three_hit_unaligned[$a]}++;
				$unaligned{$query_name}{three_chimera_count}++;
			}
			elsif ($three_query_unaligned[$a] - $three_hit_unaligned[$a] >= $options{min_3_trimming_length}){
				$unaligned{$query_name}{three_long_unaligned_freqs}{$three_query_unaligned[$a]}++;
				$unaligned{$query_name}{three_long_count}++;
			}
			elsif ($three_hit_unaligned[$a] - $three_query_unaligned[$a] >= $options{min_3_trimming_length}){
				$unaligned{$query_name}{three_short_unaligned_freqs}{$three_hit_unaligned[$a]}++;
				$unaligned{$query_name}{three_short_count}++;
			}
		}

		# generates consensus error call for each end; keeps track of consensus for each query and tallies consensus for each query type

		if ($unaligned{$query_name}{five_chimera_count} and $unaligned{$query_name}{five_chimera_count} / $hit_counter >= 0.5){
			$consensus{$query_name}{five} = "chimera";
			$total_five_chimera++;
		}
		elsif ($unaligned{$query_name}{five_long_count} and $unaligned{$query_name}{five_long_count} / $hit_counter >= 0.5){
			$consensus{$query_name}{five} = "long";
			$total_five_long++;
		}
		elsif ($unaligned{$query_name}{five_short_count} and $unaligned{$query_name}{five_short_count} / $hit_counter >= 0.5){
			$consensus{$query_name}{five} = "short";
			$total_five_short++;
		}

		if ($unaligned{$query_name}{three_chimera_count} and $unaligned{$query_name}{three_chimera_count} / $hit_counter >= 0.5){
			$consensus{$query_name}{three} = "chimera";
			$total_three_chimera++;
		}
		elsif ($unaligned{$query_name}{three_long_count} and $unaligned{$query_name}{three_long_count} / $hit_counter >= 0.5){
			$consensus{$query_name}{three} = "long";
			$total_three_long++;
		}
		elsif ($unaligned{$query_name}{three_short_count} and $unaligned{$query_name}{three_short_count} / $hit_counter >= 0.5){
			$consensus{$query_name}{three} = "short";
			$total_three_short++;
		}
	}

	# if short >= 40% and 5' long >= 40% bias towards long

	my $five_ambiguous = my $three_ambiguous = "N";
	if ($total_five_long and $total_five_short and $total_five_long / (scalar @individual_faas - 2 - $no_hits) >= 0.4 and 
		$total_five_short / (scalar @individual_faas - 2 - $no_hits) >= 0.4){ # minus 2 to account for ./ and ../
		$five_ambiguous = "Y";
	}
	if ($total_three_long and $total_three_short and $total_three_long / (scalar @individual_faas - 2 - $no_hits) >= 0.4 and 
		$total_three_short / (scalar @individual_faas - 2 - $no_hits) >= 0.4){ # minus 2 to account for ./ and ../
		$three_ambiguous = "Y";
	}
	my $start_position = my $add_5Xs = my $add_3Xs = 0;

	# read through each error-containing query again, deciding whether or not to make corrections in each case taking whether or not the 
	# long/short decision is ambiguous or not

	foreach my $individual_faa (@individual_faas){
		next if ($individual_faa =~ /^\./);
		$individual_faa =~ /((\w|-)+)\..*$/ or die "Cannot match name: $individual_faa";
		my $query_name = $1;

		# read query fasta file

		my @sequence = ();
		my $header;
		open (INFILE, "ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa") 
			or die "Cannot open ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa";
		while (<INFILE>){
			chomp;
			if (/>/){
				$header = $_;
			}
			else {
				@sequence = (@sequence, split '');
			}
		}
		close INFILE;
		my $query_length = scalar @sequence;

		# read nucleotide file, if relevant

		my @nucl_sequence = ();
		my $nucl_header;
		if ($options{protein} eq "N"){
			open (INFILE, "ORFcor_individual_sequence_files_nucl/$input_file_names[$counter]/$query_name.ffn") or die 
				"Cannot open ORFcor_individual_sequence_files_nucl/$input_file_names[$counter]/$query_name.ffn";
			while (<INFILE>){
				chomp;
				if (/>/){
					$nucl_header = $_;
				}
				else {
					@nucl_sequence = (@nucl_sequence, split '');
				}
			}
			close INFILE;
		}


		my $end_position = $#sequence;
		my $start_position = my $add_5Xs = my $add_3Xs = my $five_chimera = my $five_long = my $five_short = my $three_chimera = 
			my $three_long = my $three_short = 0;
		my $consensus_five_chimera_query_unaligned = my $consensus_five_chimera_hit_unaligned = my $consensus_five_query_unaligned = 
			my $consensus_five_hit_unaligned = my $consensus_three_chimera_query_unaligned = my $consensus_three_chimera_hit_unaligned =
			my $consensus_three_query_unaligned = my $consensus_three_hit_unaligned = "N";

		# 5' chimera

		if ($consensus{$query_name}{five} and $consensus{$query_name}{five} eq "chimera"){

			# finds candidate consensus query starts and stops, i.e., those having >= 33% abundance, and choses the longest

			foreach (sort keys %{$unaligned{$query_name}{five_chimera_query_unaligned_freqs}}){
				if ($consensus_five_chimera_query_unaligned eq "N" and 
					$unaligned{$query_name}{five_chimera_query_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{five_chimera_count} >= 0.33){
					$consensus_five_chimera_query_unaligned = $_;
				}
				elsif ($consensus_five_chimera_query_unaligned ne "N" and 
					$unaligned{$query_name}{five_chimera_query_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{five_chimera_count} >= 0.33 and $_ > $consensus_five_chimera_query_unaligned){
					$consensus_five_chimera_query_unaligned = $_;
				}
			}

			# if no query consensus found, uses average starting position

			if ($consensus_five_chimera_query_unaligned eq "N"){
				my $total = 0;
				foreach (sort keys %{$unaligned{$query_name}{five_chimera_query_unaligned_freqs}}){
					$total += $unaligned{$query_name}{five_chimera_query_unaligned_freqs}{$_} * $_;
				}
				$consensus_five_chimera_query_unaligned = $total / $unaligned{$query_name}{five_chimera_count};
			}

			# finds candidate consensus hit starts and stops, i.e., those having >= 33% abundance, and choses the longest

			foreach (sort keys %{$unaligned{$query_name}{five_chimera_hit_unaligned_freqs}}){
				if ($consensus_five_chimera_hit_unaligned eq "N" and $unaligned{$query_name}{five_chimera_hit_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{five_chimera_count} >= 0.33){
					$consensus_five_chimera_hit_unaligned = $_;
				}
				elsif ($consensus_five_chimera_hit_unaligned ne "N" and 
					$unaligned{$query_name}{five_chimera_hit_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{five_chimera_count} >= 0.33 and $_ > $consensus_five_chimera_hit_unaligned){
					$consensus_five_chimera_hit_unaligned = $_;
				}
			}

			# if no hit consensus found, uses average starting position

			if ($consensus_five_chimera_hit_unaligned eq "N"){
				my $total = 0;
				foreach (sort keys %{$unaligned{$query_name}{five_chimera_hit_unaligned_freqs}}){
					$total += $unaligned{$query_name}{five_chimera_hit_unaligned_freqs}{$_} * $_;
				}
				$consensus_five_chimera_hit_unaligned = $total / $unaligned{$query_name}{five_chimera_count};
			}

			$start_position = $consensus_five_chimera_query_unaligned;
			$add_5Xs = $consensus_five_chimera_hit_unaligned;
			$five_chimera = $consensus_five_chimera_query_unaligned;
		}

		# 5' short or ambiguous

		elsif ($consensus{$query_name}{five} and $consensus{$query_name}{five} eq "short" or $five_ambiguous eq "Y"){

			if ($unaligned{$query_name}{five_short_unaligned_freqs}){ # if $five_ambiguous eq "Y" seq may not be short

				# finds candidate consensus starts and stops, i.e., those having >= 33% abundance, and choses the longest

				foreach (sort keys %{$unaligned{$query_name}{five_short_unaligned_freqs}}){
					if ($consensus_five_hit_unaligned eq "N" and $unaligned{$query_name}{five_short_unaligned_freqs}{$_} / 
						$unaligned{$query_name}{five_short_count} >= 0.33){
						$consensus_five_hit_unaligned = $_;
					}
					elsif ($consensus_five_hit_unaligned ne "N" and $unaligned{$query_name}{five_short_unaligned_freqs}{$_} / 
						$unaligned{$query_name}{five_short_count} >= 0.33 and $_ > $consensus_five_hit_unaligned){
						$consensus_five_hit_unaligned = $_;
					}
				}
	
				# if no consensus found, uses average starting position
	
				if ($consensus_five_hit_unaligned eq "N"){
					my $total = 0;
					foreach (sort keys %{$unaligned{$query_name}{five_short_unaligned_freqs}}){
						$total += $unaligned{$query_name}{five_short_unaligned_freqs}{$_} * $_;
					}
					$consensus_five_hit_unaligned = $total / $unaligned{$query_name}{five_short_count};
				}

				$add_5Xs = $consensus_five_hit_unaligned;
				$five_short = $add_5Xs;
			}
		}

		# 5' unambiguously long

		elsif ($consensus{$query_name}{five} and $consensus{$query_name}{five} eq "long" and $five_ambiguous eq "N"){

			# finds candidate consensus starts and stops, i.e., those having >= 33% abundance, and choses the longest

			foreach (sort keys %{$unaligned{$query_name}{five_long_unaligned_freqs}}){
				if ($consensus_five_query_unaligned eq "N" and $unaligned{$query_name}{five_long_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{five_long_count} >= 0.33){
					$consensus_five_query_unaligned = $_;
				}
				elsif ($consensus_five_query_unaligned ne "N" and $unaligned{$query_name}{five_long_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{five_long_count} >= 0.33 and $_ < $consensus_five_query_unaligned){
					$consensus_five_query_unaligned = $_;
				}
			}

			# if no consensus found, uses average starting position

			if ($consensus_five_query_unaligned eq "N"){
				my $total = 0;
				foreach (sort keys %{$unaligned{$query_name}{five_long_unaligned_freqs}}){
					$total += $unaligned{$query_name}{five_long_unaligned_freqs}{$_} * $_;
				}
				$consensus_five_query_unaligned = $total / $unaligned{$query_name}{five_long_count};
			}

			$start_position = $consensus_five_query_unaligned;
			$five_long = $consensus_five_query_unaligned;
		}

		# 3' chimera

		if ($consensus{$query_name}{three} and $consensus{$query_name}{three} eq "chimera"){

			# finds candidate consensus query starts and stops, i.e., those having >= 33% abundance, and choses the longest

			foreach (sort keys %{$unaligned{$query_name}{three_chimera_query_unaligned_freqs}}){
				if ($consensus_three_chimera_query_unaligned eq "N" and 
					$unaligned{$query_name}{three_chimera_query_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{three_chimera_count} >= 0.33){
					$consensus_three_chimera_query_unaligned = $_;
				}
				elsif ($consensus_three_chimera_query_unaligned ne "N" 
					and $unaligned{$query_name}{three_chimera_query_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{three_chimera_count} >= 0.33 and $_ > $consensus_three_chimera_query_unaligned){
					$consensus_three_chimera_query_unaligned = $_;
				}
			}

			# if no query consensus found, uses average starting position

			if ($consensus_three_chimera_query_unaligned eq "N"){
				my $total = 0;
				foreach (sort keys %{$unaligned{$query_name}{three_chimera_query_unaligned_freqs}}){
					$total += $unaligned{$query_name}{three_chimera_query_unaligned_freqs}{$_} * $_;
				}
				$consensus_three_chimera_query_unaligned = $total / $unaligned{$query_name}{three_chimera_count};
			}

			# finds candidate consensus hit starts and stops, i.e., those having >= 33% abundance, and choses the longest

			foreach (sort keys %{$unaligned{$query_name}{three_chimera_hit_unaligned_freqs}}){
				if ($consensus_three_chimera_hit_unaligned eq "N" and 
					$unaligned{$query_name}{three_chimera_hit_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{three_chimera_count} >= 0.33){
					$consensus_three_chimera_hit_unaligned = $_;
				}
				elsif ($consensus_three_chimera_hit_unaligned ne "N" and 
					$unaligned{$query_name}{three_chimera_hit_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{three_chimera_count} >= 0.33 and $_ > $consensus_three_chimera_hit_unaligned){
					$consensus_three_chimera_hit_unaligned = $_;
				}
			}

			# if no hit consensus found, uses average starting position

			if ($consensus_three_chimera_hit_unaligned eq "N"){
				my $total = 0;
				foreach (sort keys %{$unaligned{$query_name}{three_chimera_hit_unaligned_freqs}}){
					$total += $unaligned{$query_name}{three_chimera_hit_unaligned_freqs}{$_} * $_;
				}
				$consensus_three_chimera_hit_unaligned = $total / $unaligned{$query_name}{three_chimera_count};
			}

			$end_position = $query_length - $consensus_three_chimera_query_unaligned - 1;
			$add_3Xs = $consensus_three_chimera_hit_unaligned;
			$three_chimera = $consensus_three_chimera_query_unaligned;
		}

		# 3' short or ambiguous

		elsif ($consensus{$query_name}{three} and $consensus{$query_name}{three} eq "short" or $three_ambiguous eq "Y"){

			if ($unaligned{$query_name}{three_short_unaligned_freqs}){ # if $three_ambiguous eq "Y" seq may not be short

				# finds candidate consensus starts and stops, i.e., those having >= 33% abundance, and choses the longest

				foreach (sort keys %{$unaligned{$query_name}{three_short_unaligned_freqs}}){
					if ($consensus_three_hit_unaligned eq "N" and $unaligned{$query_name}{three_short_unaligned_freqs}{$_} / 
						$unaligned{$query_name}{three_short_count} >= 0.33){
						$consensus_three_hit_unaligned = $_;
					}
					elsif ($consensus_three_hit_unaligned ne "N" and $unaligned{$query_name}{three_short_unaligned_freqs}{$_} / 
						$unaligned{$query_name}{three_short_count} >= 0.33 and $_ > $consensus_three_hit_unaligned){
						$consensus_three_hit_unaligned = $_;
					}
				}

				# if no consensus found, uses average starting position

				if ($consensus_three_hit_unaligned eq "N"){
					my $total = 0;
					foreach (sort keys %{$unaligned{$query_name}{three_short_unaligned_freqs}}){
						$total += $unaligned{$query_name}{three_short_unaligned_freqs}{$_} * $_;
					}
					$consensus_three_hit_unaligned = $total / $unaligned{$query_name}{three_short_count};
				}
				$add_3Xs = $consensus_three_hit_unaligned;
				$three_short = $add_3Xs;
			}
		}

		# 3' unambiguously long

		elsif ($consensus{$query_name}{three} and $consensus{$query_name}{three} eq "long" and $three_ambiguous eq "N"){

			# finds candidate consensus starts and stops, i.e., those having >= 33% abundance, and choses the longest

			foreach (sort keys %{$unaligned{$query_name}{three_long_unaligned_freqs}}){
				if ($consensus_three_query_unaligned eq "N" and $unaligned{$query_name}{three_long_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{three_long_count} >= 0.33){
					$consensus_three_query_unaligned = $_;
				}
				elsif ($consensus_three_query_unaligned ne "N" and $unaligned{$query_name}{three_long_unaligned_freqs}{$_} / 
					$unaligned{$query_name}{three_long_count} >= 0.33 and $_ < $consensus_three_query_unaligned){
					$consensus_three_query_unaligned = $_;
				}
			}

			# if no consensus found, uses average starting position

			if ($consensus_three_query_unaligned eq "N"){
				my $total = 0;
				foreach (sort keys %{$unaligned{$query_name}{three_long_unaligned_freqs}}){
					$total += $unaligned{$query_name}{three_long_unaligned_freqs}{$_} * $_;
				}
				$consensus_three_query_unaligned = $total / $unaligned{$query_name}{three_long_count};
			}
			$end_position = $query_length - $consensus_three_query_unaligned - 1;
			$three_long = $consensus_three_query_unaligned;
		}			

		# alter protein sequences

		@sequence = @sequence[$start_position..$end_position]; # trim query sequence
	
		for (0..$add_5Xs - 1){
			unshift @sequence, "X";
		}
		for (0..$add_3Xs - 1){
			push @sequence, "X";
		}

		# adjust nucleotide sequence as necessary

		if ($options{protein} eq "N"){
			@nucl_sequence = @nucl_sequence[($start_position * 3)..($end_position * 3 + 2)];

			for (0..($add_5Xs - 1) * 3 + 2){
				unshift @nucl_sequence, "N";
			}
			for (0..($add_3Xs - 1) * 3 + 2){
				push @nucl_sequence, "N";
			}		
		}					

		push @return, "$query_name,$five_chimera,$five_long,$five_short,$three_chimera,$three_long,$three_short";		

		open (OUTFILE, ">ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa.corrected") 
			or die "Cannot open ORFcor_individual_sequence_files/$input_file_names[$counter]/$individual_faa.corrected";
		print OUTFILE $header, "\n", @sequence, "\n";
		close OUTFILE;

		if ($options{protein} eq "N"){
			open (OUTFILE, ">ORFcor_individual_sequence_files_nucl/$input_file_names[$counter]/$query_name.ffn.corrected") 
				or die "Cannot open ORFcor_individual_sequence_files_nucl/$input_file_names[$counter]/$query_name.ffn.corrected";
			print OUTFILE $nucl_header, "\n", @nucl_sequence, "\n";
			close OUTFILE;
		}
			
	}

	# cleans up BLAST db

	system ("rm -f $input_file_list[$counter].*");
	system ("rm -f *makeblastdb.log");

	# concatenates corrected .faa files into a single file per hmm

	system ("cat ORFcor_individual_sequence_files/$input_file_names[$counter]/*.corrected > ORFcor_multiple_sequence_files/$input_file_names[$counter]_corrected.faa");
	if ($options{protein} eq "N"){
		system ("cat ORFcor_individual_sequence_files_nucl/$input_file_names[$counter]/*.corrected > ORFcor_multiple_sequence_files_nucl/$input_file_names[$counter]_corrected.ffn");
	}

	return @return;
}
